/*
Copyright (c) 2010-2011 The Regents of the University of California. All rights
reserved.

Permission is hereby granted, without written agreement and without license or 
royalty fees, to use, copy, modify, and distribute this software and its 
documentation for any purpose, provided that the above copyright notice and the
following two paragraphs appear in all copies of this software.

IN NO EVENT SHALL THE UNIVERSITY OF CALIFORNIA BE LIABLE TO ANY PARTY FOR
DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES ARISING OUT OF
THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF THE UNIVERSITY OF 
CALIFORNIA HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

THE UNIVERSITY OF CALIFORNIA SPECIFICALLY DISCLAIMS ANY WARRANTIES, INCLUDING,
BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR 
A PARTICULAR PURPOSE. THE SOFTWARE PROVIDED HEREUNDER IS ON AN "AS IS" BASIS, 
AND THE UNIVERSITY OF CALIFORNIA HAS NO OBLIGATION TO PROVIDE MAINTENANCE, 
SUPPORT, UPDATES, ENHANCEMENTS, OR MODIFICATIONS.
*/

/*
 * @file main.c
 * Example program that demonstrates how to interface the PTP protocol with the
 * uIP protocol stack @link {http://www.sics.se/~adam/uip/index.php/Main_Page} 
 * and run the PTP protocol on the SH7216 Demonstration Kit containing the 
 * DP83640 Precision PHYTER. Program is just for demonstration.
 *
 * @author Michael Zimmer (mzimmer@eecs.berkeley.edu)
 * @version $Id: main.c 71 2011-05-28 00:35:19Z mzimmer $
 */

#include "DP83640.h"
#include "typedefine.h"
#include "uip.h"
#include "uip_arp.h"
#include "eth.h"
#include <string.h>
#include "main.h"

// uIP connections for multicast PTP event and general messages.
struct uip_udp_conn * ptp_event_conn;
struct uip_udp_conn * ptp_general_conn;

// An event queue allows the PTP actor to output events which 
// fire the PTP actor at a later time.
// This is used for timeouts.
EventInList *eventQueue;

int main(void)
{
	char i;
	// For LCD screen.
	char cmd_clr_scr[] = {27,'[','2','J',0};
	char cmd_cur_home[] = {27,'[','H',0};
	
	// For uIP connection.
	uip_ipaddr_t IPAddress = SRC_IP;
	struct uip_eth_addr MACAddress = SRC_MAC;

	// Initialize LCD.
	InitialiseDisplay(); 

	// Initialize ethernet. 
	while (R_ETHER_ERROR == R_Ether_Open(0, &MACAddress.addr[0]));

	// Initialize and configure uIP.
	uip_sethostaddr(IPAddress);
	uip_setethaddr(MACAddress);
	uip_init();

	// Initialize PTP.
	ptp_init();

	// Main operation.
	while (1)
	{
    	// Check if PTP message has been received.
		uip_len = R_Ether_Read(0, (void *)uip_buf);
    	if (uip_len > 0)
    	{		
			uip_input();
			// If calling uip_input() results in data to be sent
			// over the network, uip_len will be > 0.
        	if (uip_len > 0)
        	{
				// Add correct ethernet header for multicast PTP.
				add_ptp_hdr();
				// Write to network.
          		R_Ether_Write(0, (void *)uip_buf, (uint32_t)uip_len);
        	}
    	}
		else
    	{
			// Periodically check event queue.
			for (i = 0; i < UIP_UDP_CONNS; i++)
     		{
        		uip_udp_periodic(i);
        		// If calling uip_upd_periodic() results in data to be sent
				// over the network, uip_len will be > 0.
        		if (uip_len > 0)
        		{
					// Add correct ethernet header for PTP.
		  			add_ptp_hdr();
          			// Write to network.
          			R_Ether_Write(0, (void *)uip_buf, (uint32_t)uip_len);
        		}
      		}
    	}	
	}
	return 0;
}

// Do nothing.
void uip_appcall(void)
{
	
}

// Called whenever new UDP message has arrived or for periodic polling.
// Responsible for firing the PTPActor with the PTP message, or with an
// event from the event queue when the timestamp has expired.
void udp_appcall(void)
{
	
	struct uip_udpip_hdr* udp_header =
            (struct uip_udpip_hdr *)&uip_buf[UIP_LLH_LEN];
	
	if(uip_newdata() && (udp_header->destport == PTP_EVENT_PORT || 
                udp_header->destport == PTP_GENERAL_PORT))
	{
		Event eventIn, eventOut;
		// Build event.
		eventIn.eventType = EVENT_MSG_IN;
		eventIn.eventDataPointer = (UInteger8 *)uip_appdata;
		// Fire PTPActor.
		PTPActor(&eventIn);
	}
	else
	{
		// Check event queue.
		checkEventQueue();	
	}
}

// Hardware initialization for PTP.
void ptp_init()
{
  	UInteger16 reg;
	Event initEvent;
	uip_ipaddr_t ip = MULTICAST_IP;
  
	// PTP Event message connection.
	ptp_event_conn = uip_udp_new(&ip, HTONS(PTP_EVENT_PORT));
	if(ptp_event_conn != NULL) {
		uip_udp_bind( ptp_event_conn , HTONS(LOCAL_EVENT_PORT));
	}
	
	// PTP General message connection.
	ptp_general_conn = uip_udp_new(&ip, HTONS(PTP_GENERAL_PORT));
	if(ptp_general_conn != NULL) {
		uip_udp_bind( ptp_general_conn , HTONS(LOCAL_GENERAL_PORT));
	}
	
	// Enable PTP clock.
	_phy_write(PAGE_SEL_REG, PTP_CTL_PAGE);
	_phy_write(PTP_CTL_REG, ( 
		PTP_CTL_PTP_ENABLE
	));
	
	// TX Configuration 0
	_phy_write(PAGE_SEL_REG, PTP_TXCFG0_PAGE);
	_phy_write(PTP_TXCFG0_REG, (
		#if TWOSTEPFLAG == FALSE
		PTP_TXCFG0_SYNC_1STEP | 
		#endif
		//PTP_TXCFG0_DR_INSERT | // TODO
		//PTP_TXCFG0_IGNORE_2STEP | 
		PTP_TXCFG0_CRC_1STEP | // TODO
		PTP_TXCFG0_CHK_1STEP | // TODO
		PTP_TXCFG0_IP1588_EN |
		PTP_TXCFG0_TX_IPV4_EN |
		(0x2 << PTP_RXCFG0_RX_PTP_VER_SHIFT) | 
		PTP_TXCFG0_TX_TS_EN
	));

	// RX Configuration 0
	_phy_write(PAGE_SEL_REG, PTP_RXCFG0_PAGE);
	_phy_write(PTP_RXCFG0_REG, (
		//PTP_RXCFG0_DOMAIN_EN |
		//PTP_RXCFG0_ALT_MAST_DIS |
		//PTP_RXCFG0_USER_IP_SEL |
		//PTP_RXCFG0_USER_IP_EN |
		//PTP_RXCFG0_RX_SLAVE |
		PTP_RXCFG0_IP1588_EN2 |
		PTP_RXCFG0_IP1588_EN1 |
		PTP_RXCFG0_IP1588_EN0 |
		PTP_RXCFG0_RX_IPV4_EN |
		(0x2 << PTP_RXCFG0_RX_PTP_VER_SHIFT) |
		PTP_RXCFG0_RX_TS_EN 
	));
	
	
	// RX Configuration 3
	_phy_write(PAGE_SEL_REG, PTP_RXCFG3_PAGE);
	_phy_write(PTP_RXCFG3_REG, (
		(0xC << PTP_RXCFG3_TS_MIN_IFG_SHIFT) |  
		PTP_RXCFG3_ACC_UDP |       //TODO 
		PTP_RXCFG3_ACC_CRC        //TODO   
		//PTP_RXCFG3_TS_APPEND |         
		//PTP_RXCFG3_TS_INSERT        
		//PTP_RXCFG3_PTP_DOMAIN_SHIFT   
		//PTP_RXCFG3_PTP_DOMAIN_MASK    
	));
	
	
	// RX Configuration 4
	_phy_write(PAGE_SEL_REG, PTP_RXCFG4_PAGE);
	_phy_write(PTP_RXCFG4_REG, (
		 //PTP_IPV4_UDP_MOD |
		 //PTP_TS_SEC_EN |
		 //(0x3 << PTP_TS_SEC_LEN_SHIFT)  
		// PTP_RXTS_NS_OFF_SHIFT 
		// PTP_RXTS_NS_OFF_MASK  
		// PTP_RXTS_SEC_OFF_SHIFT 
		// PTP_RXTS_SEC_OFF_MASK
		0  
	));
	
	// Trigger 0 as pulse per seconds
	_phy_write(PAGE_SEL_REG, PTP_TRIG_PAGE);
	_phy_write(PTP_TRIG_REG, (
		//PTP_TRIG_PULSE |
		PTP_TRIG_PER |
		//PTP_TRIG_IF_LATE | 
		//PTP_TRIG_NOTIFY | 
		(0x8 << PTP_TRIG_GPIO_SHIFT) |
		//PTP_TRIG_TOGGLE
		(0x0 << PTP_TRIG_CSEL_SHIFT) |
		PTP_TRIG_WR          
	));
	
	_phy_write(PAGE_SEL_REG, PTP_CTL_PAGE);
	_phy_write(PTP_CTL_REG, ( 
		(0x0 << PTP_CTL_TRIG_SEL_SHIFT) |
		PTP_CTL_TRIG_LOAD
	));
	_phy_write(PTP_TDR_REG,0x0); //expire ns lower
	_phy_write(PTP_TDR_REG,0x0); //expire ns upper
	_phy_write(PTP_TDR_REG,0x1); //expire s lower
	_phy_write(PTP_TDR_REG,0x0); //expire s upper
	_phy_write(PTP_TDR_REG,0x6500); //pulse width ns lower
	_phy_write(PTP_TDR_REG,0x1DCD); //pulse width ns upper
	_phy_write(PTP_TDR_REG,0x0); //
	_phy_write(PTP_TDR_REG,0x0); //
	_phy_write(PTP_CTL_REG, ( 
		(0x0 << PTP_CTL_TRIG_SEL_SHIFT) |
		PTP_CTL_TRIG_EN
	));
	
	// Add PowerUp Event.
	initEvent.eventType = EVENT_POWERUP;
	getClock(&initEvent.eventTimestamp);
	addToEventQueue(&initEvent);

}


void add_ptp_hdr(void)
{
  
   // Multicast
   struct uip_eth_addr dest_mac = DEST_MAC;
   struct uip_eth_addr my_mac = SRC_MAC;
   struct uip_eth_hdr *header = (struct uip_eth_hdr *)&uip_buf[0];

  memcpy(&header->dest.addr, &dest_mac, 6);
  memcpy(&header->src.addr, &my_mac, 6);  
  header->type = HTONS(UIP_ETHTYPE_IP);
  uip_len += sizeof(struct uip_eth_hdr);
  
}


// Compare two timestamps. Returns TRUE if x < y.
UInteger8 timestampLess(const Timestamp *x, const Timestamp *y) {
	if(x->secondsField.msb < y->secondsField.msb)
		return TRUE;
	else if(x->secondsField.msb > y->secondsField.msb)
		return FALSE;
	else if(x->secondsField.lsb < y->secondsField.lsb)
		return TRUE;
	else if(x->secondsField.lsb > y->secondsField.lsb)
		return FALSE;
	else if(x->nanosecondsField < y->nanosecondsField)
		return TRUE;
	else
		return 0;	
}

void checkEventQueue() {
	if(eventQueue != NULL) {

        EventInList *toFree;
		Timestamp timestamp;
		getClock(&timestamp);
		// Check timestamp of first event in queue, and fire actor if needed.
		if(!timestampLess(&timestamp,&eventQueue->event.eventTimestamp)) {
            
            toFree = eventQueue;

            // Update event queue.
			eventQueue = eventQueue->nextEventInList;

            // Fire actor with event.
			PTPActor(&(toFree->event));
			
            // Remove from event queue.
			free(toFree);

        }
	}
}

// Add an event to the event queue (maintain timestamp order).  
// Event queue is implemented as a singly linked list.
void addToEventQueue(Event *event) {

	EventInList *newEventInList = (EventInList *)malloc(sizeof(EventInList));
	newEventInList->event = *event;
	// If the event queue is empty.
	if(eventQueue == NULL) {
		eventQueue = newEventInList;
		newEventInList->nextEventInList = NULL;
	} 
	// If event queue is not empty, put in timestamp order.	
	else {
		EventInList *prev = eventQueue;
		EventInList *curr = eventQueue->nextEventInList;
		// Check if new event needs to be put at front of the event queue.
		if(timestampLess(&newEventInList->event.eventTimestamp,
                    &eventQueue->event.eventTimestamp)) {
			newEventInList->nextEventInList = eventQueue;
			eventQueue = newEventInList;	
		}
		else {
			// Check if new event needs to be put in middle of the event queue.
			while(curr != NULL) {
				if(timestampLess(&newEventInList->event.eventTimestamp,
                            &curr->event.eventTimestamp)) {
					prev->nextEventInList = newEventInList;
					newEventInList->nextEventInList = curr;	
					break;
				}
				prev = curr;
				curr = curr->nextEventInList;
			}
			// Check if new event needs to be put at end of the event queue.
			if(curr == NULL) {
				prev->nextEventInList = newEventInList;
				newEventInList->nextEventInList = NULL;
			}
		}
	}
}
	
void fixTimestampsInEventQueue(const TimeInterval *timeInterval) {
	
	EventInList *currEvent;
		// Fix timestamps in event queue.
	currEvent = eventQueue;
	while(currEvent != NULL) {
		addTimeIntervalToTimestamp(timeInterval,
                &currEvent->event.eventTimestamp);
		currEvent = currEvent->nextEventInList;	
	}
	
	
}







