/*
Copyright (c) 2010-2011 The Regents of the University of California. All rights
reserved.

Permission is hereby granted, without written agreement and without license or 
royalty fees, to use, copy, modify, and distribute this software and its 
documentation for any purpose, provided that the above copyright notice and the
following two paragraphs appear in all copies of this software.

IN NO EVENT SHALL THE UNIVERSITY OF CALIFORNIA BE LIABLE TO ANY PARTY FOR
DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES ARISING OUT OF
THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF THE UNIVERSITY OF 
CALIFORNIA HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

THE UNIVERSITY OF CALIFORNIA SPECIFICALLY DISCLAIMS ANY WARRANTIES, INCLUDING,
BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR 
A PARTICULAR PURPOSE. THE SOFTWARE PROVIDED HEREUNDER IS ON AN "AS IS" BASIS, 
AND THE UNIVERSITY OF CALIFORNIA HAS NO OBLIGATION TO PROVIDE MAINTENANCE, 
SUPPORT, UPDATES, ENHANCEMENTS, OR MODIFICATIONS.
*/

/**
 * @file ptp-clock.c
 * Define the platform-dependent interface with the real time clock that is
 * required for the correct operation of PTP.
 *
 * @author Michael Zimmer (mzimmer@eecs.berkeley.edu)
 * @version $Id: ptp-clock.c 51 2011-04-12 01:19:19Z mzimmer $
 */

#include "../ptp-protocol.h"
#include "DP83640.h"
 
/** Attempt to provide an accurate clock timestamp. In
 * this implementation, the clock timestamp is retrieved by reading the 
 * timestamp from the registers on the DP83640 via a serial interface.
 */
void getClock(Timestamp *timestamp) {
   
    UInteger16 l, h;
    
    // Read clock value from time data register (DP83640, 14.5.2).
    _phy_write(PAGE_SEL_REG, PTP_CTL_PAGE);
    _phy_write(PTP_CTL_REG, PTP_CTL_PTP_RD_CLK);
    
    _phy_write(PAGE_SEL_REG, PTP_TDR_PAGE);
    l = _phy_read(PTP_TDR_REG);
    h = _phy_read(PTP_TDR_REG);
    timestamp->nanosecondsField = (h << 16) | l;
    
    l = _phy_read(PTP_TDR_REG);
    h = _phy_read(PTP_TDR_REG);
    timestamp->secondsField.lsb = (h << 16) | l;
    timestamp->secondsField.msb = 0;

}

/** Step the clock by -timeInterval.  In this implementation,
 * this is done by setting registers on the DP83640 via a serial interface.
 * Note: The timeInterval supplied is an offset (slave time - master time)
 * so actual adjustment is the opposite sign of the offset.
 * The timestamps of events in the event queue are also fixed to account for 
 * clock step.
 */
void stepClockTimeInterval(const TimeInterval *timeInterval) {

	TimeInterval t;
    Integer32 s, ns;
    Integer16 ns_l,ns_h,s_l,s_h;
	Timestamp temp;
    
    // Put correct values into time data register (DP83640, 14.5.2).
	t.scaledNanoseconds = -timeInterval->scaledNanoseconds;
	fixTimestampsInEventQueue(&t);
	// Unscale.
	t.scaledNanoseconds >>= 16;
	// Split into seconds and nanoseconds.
	//TODO: overflow?
    s = t.scaledNanoseconds / (1000000000);
    ns = t.scaledNanoseconds % (1000000000);
        
    ns_l = (0xFFFF & ns);
    ns_h = (ns >> 16);
    s_l = (0xFFFF & s);
    s_h = (s >> 16);
        
    _phy_write(PAGE_SEL_REG, PTP_TDR_PAGE);
    _phy_write(PTP_TDR_REG, ns_l);
    _phy_write(PTP_TDR_REG, ns_h);
    _phy_write(PTP_TDR_REG, s_l);
    _phy_write(PTP_TDR_REG, s_h);
    
    // Step clock to new value by modifying PTP control register
    // (DP83640, 14.5.1).
    _phy_write(PAGE_SEL_REG, PTP_CTL_PAGE);
    _phy_write(PTP_CTL_REG, PTP_CTL_PTP_STEP_CLK);

    fixTimestampsInEventQueue(&t);
    
}

/** Adjust the rate of the clock by rateAdj (in units of 
 * 100ps/sec).  In this implementation, this is done by correctly
 * setting registers in the DP83640 via a serial interface. 
 * Positive is higher frequency.
 */
void rateAdjustClock(Integer32 rateAdj) {

    Integer16 l,h;
    Integer64 r;
	 
    // Convert to units of 2^-32 ns per clock cycle.
    r = rateAdj;
    r = (r << 32) / (10 * PTP_CLK_RATE);
    rateAdj = r;
    //TODO: overflow?
    
    // Write correct values to PTP high and low rate registers
    // (DP83640, 14.5.5, 14.5.6).
    _phy_write(PAGE_SEL_REG, PTP_RATEH_PAGE);
    if(rateAdj > 0) {
        h = (0x03FF & (rateAdj >> 16));  
    } 
    else {
        rateAdj = -rateAdj;
        h = 0x8000 | (0x03FF & (rateAdj >> 16));
	}
	_phy_write(PTP_RATEH_REG, h);
	l = 0xFFFF & rateAdj;
    _phy_write(PTP_RATEL_REG, l);

}

/** Compute: timeInterval = t2 - t1.
 * This function may be implementation specific depending on how Integer64
 * values are handled in computations.
 */ 
void toTimeInterval(const Timestamp *t1, const Timestamp *t2, 
        TimeInterval *timeInterval) {    

    // Currently ignores msb of seconds.
    timeInterval->scaledNanoseconds = t2->secondsField.lsb;
    timeInterval->scaledNanoseconds -= t1->secondsField.lsb;
    timeInterval->scaledNanoseconds *= (1000000000);
    timeInterval->scaledNanoseconds += t2->nanosecondsField;
    timeInterval->scaledNanoseconds -= t1->nanosecondsField;
    // Scale.
    timeInterval->scaledNanoseconds <<= 16;	

}

/** Compute: timeInterval = timeInterval + timestamp.
 * This function may be implementation specific depending on how Integer64
 * values are handled in computations.
 */ 
void addTimeIntervalToTimestamp(const TimeInterval *timeInterval, 
        Timestamp *timestamp) {
    
    Integer64 t;
    Integer32 s;
    Integer32 ns;
   
    // Convert to nanoseconds. (Timestamp cannot handle fractional nanoseconds)
    t = timeInterval->scaledNanoseconds >> 16;
    
	// TODO: overflow?
	
    // Preliminary adjustment of seconds field.
    s = timestamp->secondsField.lsb + (t / 1000000000);
    
    // Adjustment of nanoseconds field.
    ns = timestamp->nanosecondsField + (t % 1000000000);
    
    // Consider overflow of nanoseconds field.
    if(ns > 1000000000) {
        s++;
        ns -= 1000000000;    
    }
    // Consider underflow of nanoseconds field.
    if(ns < 0) {
        s--;
        ns += 1000000000;    
    }
    timestamp->secondsField.lsb = s;
    timestamp->nanosecondsField = ns;
    
}

/** Result in PTPActor() being fired with this event at the
 * time event.eventTimestamp.  This is implementation specific because it is
 * the responsibily of the program utilizing the PTP protocol to fire the 
 * PTPActor() at the correct time. 
 */
void addFutureEvent(const Event *event) {
    
    addToEventQueue(event);

}


