/*
Copyright (c) 2010-2011 The Regents of the University of California. All rights
reserved.

Permission is hereby granted, without written agreement and without license or 
royalty fees, to use, copy, modify, and distribute this software and its 
documentation for any purpose, provided that the above copyright notice and the
following two paragraphs appear in all copies of this software.

IN NO EVENT SHALL THE UNIVERSITY OF CALIFORNIA BE LIABLE TO ANY PARTY FOR
DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES ARISING OUT OF
THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF THE UNIVERSITY OF 
CALIFORNIA HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

THE UNIVERSITY OF CALIFORNIA SPECIFICALLY DISCLAIMS ANY WARRANTIES, INCLUDING,
BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR 
A PARTICULAR PURPOSE. THE SOFTWARE PROVIDED HEREUNDER IS ON AN "AS IS" BASIS, 
AND THE UNIVERSITY OF CALIFORNIA HAS NO OBLIGATION TO PROVIDE MAINTENANCE, 
SUPPORT, UPDATES, ENHANCEMENTS, OR MODIFICATIONS.
*/

/**
 * @file ptp-msg.h
 * Define all platform-dependent functions that are required
 * to send and receive PTP messages and record their timestamps. The 
 * implementation of these functions will depend on how PTP interacts
 * with the communication network and how messages are timestamped.
 *
 * @author Michael Zimmer (mzimmer@eecs.berkeley.edu)
 * @version $Id: ptp-msg.h 51 2011-04-12 01:19:19Z mzimmer $
 */

// Send messages to the network.

/** Send an Announce message over the network, with content of message as 
 * specified in 13.5.
 * @param[in,out] ptp The current state of PTP.
 */
void sendAnnounceMsg(PTPState *ptp);

/** Send a Sync message over the network, with content of message as 
 * specified in 13.6.
 * @param[in,out] ptp The current state of PTP.
 */
void sendSyncMsg(PTPState *ptp);

/** Send a Follow_Up message over the network, with content of message as 
 * specified in 13.7.
 * @param[in,out] ptp The current state of PTP.
 */
void sendFollowUpMsg(PTPState *ptp);

/** Send a Delay_Req message over the network, with content of message as 
 * specified in 13.6.
 * @param[in,out] ptp The current state of PTP.
 */
void sendDelayReqMsg(PTPState *ptp);

/** Send a Delay_Resp message over the network, with content of message as 
 * specified in 13.8.
 * @param[in,out] ptp The current state of PTP.
 * @param[in] msg The Delay_Req message that is being responded to.
 */
void sendDelayRespMsg(PTPState *ptp, DelayReqMsg *msg);

/** Send a Management message over the network, with content of message as 
 * specified in 15.4.
 * @param[in,out] ptp The current state of PTP.
 * @param[in] msg The Management message being sent.
 */
void sendManagementMsg(PTPState *ptp, ManagementMsg *msg);


// Get timestamps.

/** Set PTPState.syncEventEgressTimestamp to the timestamp at which the latest
 * Sync message was sent.
 * @param[in,out] ptp The current state of PTP.
 */
void getSyncEventEgressTimestamp(PTPState *ptp);


/** Set PTPState.syncEventIngressTimestamp to the timestamp at which the latest
 * Sync message was received.
 * @param[in,out] ptp The current state of PTP.
 * @param[in] seqId The sequence Id of the Sync message received.
 */
Boolean getSyncEventIngressTimestamp(PTPState *ptp, UInteger16 seqId); 

/** Set PTPState.delayEventEgressTimestamp to the timestamp at which the latest
 * Delay_Req message was sent.
 * @param[in,out] ptp The current state of PTP.
 */
void getDelayEventEgressTimestamp(PTPState *ptp);

/** Set PTPState.delayEventIngressTimestamp to the timestamp at which the latest
 * Delay_Req message was received.
 * @param[in,out] ptp The current state of PTP.
 * @param[in] seqId The sequence Id of the Sync message received.
 */
Boolean getDelayEventIngressTimestamp(PTPState *ptp, UInteger16 seqId); 

// Parse received message.

/** Parse received message into provided Header data structure, but do not
 * update PTP state (13.1).
 * @param[out] header The destination for header content from message.
 */
void readHeader(Header *header);

/** Parse received Announce message into provided AnnounceMsg data structure, 
 * but do not update PTP state (13.5).
 * @param[out] msg The destination for content from Announce message.
 */
void readAnnounceMsg(AnnounceMsg *msg);

/** Parse received Sync message into provided SyncMsg data structure, 
 * but do not update PTP state (13.6).
 * @param[out] msg The destination for content from Sync message.
 */
void readSyncMsg(SyncMsg *msg);

/** Parse received Follow_Up message into provided FollowUpMsg data structure, 
 * but do not update PTP state (13.7).
 * @param[out] msg The destination for content from Follow_Up message.
 */
void readFollowUpMsg(FollowUpMsg *msg);

/** Parse received Delay_Req message into provided DelayReqMsg data structure, 
 * but do not update PTP state (13.6).
 * @param[out] msg The destination for content from Delay_Req message.
 */
void readDelayReqMsg(DelayReqMsg *msg);

/** Parse received Delay_Resp message into provided DelayRespMsg data structure,
 * but do not update PTP state (13.8).
 * @param[out] msg The destination for content from Delay_Resp message.
 */
void readDelayRespMsg(DelayRespMsg *msg);

/** Parse received Management message into provided ManagementMsg data
 * structure, but do not update PTP state (15.1).
 * @param[out] msg The destination for content from Management message.
 */
void readManagementMsg(ManagementMsg *msg);

/** Clear receive timestamp. */
void clearIngressTimestamp();


