/*
Copyright (c) 2010-2011 The Regents of the University of California. All rights
reserved.

Permission is hereby granted, without written agreement and without license or 
royalty fees, to use, copy, modify, and distribute this software and its 
documentation for any purpose, provided that the above copyright notice and the
following two paragraphs appear in all copies of this software.

IN NO EVENT SHALL THE UNIVERSITY OF CALIFORNIA BE LIABLE TO ANY PARTY FOR
DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES ARISING OUT OF
THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF THE UNIVERSITY OF 
CALIFORNIA HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

THE UNIVERSITY OF CALIFORNIA SPECIFICALLY DISCLAIMS ANY WARRANTIES, INCLUDING,
BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR 
A PARTICULAR PURPOSE. THE SOFTWARE PROVIDED HEREUNDER IS ON AN "AS IS" BASIS, 
AND THE UNIVERSITY OF CALIFORNIA HAS NO OBLIGATION TO PROVIDE MAINTENANCE, 
SUPPORT, UPDATES, ENHANCEMENTS, OR MODIFICATIONS.
*/

/**
 * @file ptp-protocol.c
 * Main operation of the IEEE 1588 Precision Time Protocol (PTP).
 * See ptp-protocol.h for more information.
 *
 * @author Michael Zimmer (mzimmer@eecs.berkeley.edu)
 * @version $Id: ptp-protocol.c 51 2011-04-12 01:19:19Z mzimmer $
 */

#include "ptp-protocol.h"
#include <stdio.h>

PTPState ptp; /**< Store state for PTP. */

UInteger16 zero = 0; /**< Zero. */

/** Handle input events by calling the corresponding event handling functions. 
 */
void PTPActor(const Event *eventIn) {

    // Handle input events.
    switch(eventIn->eventType) {
        
        case EVENT_MSG_IN:
            eventMsgIn(eventIn);
            break;
        case EVENT_POWERUP:
            eventPowerup();    
            break;
        case EVENT_QUALIFICATION_TIMEOUT_EXPIRES:
            eventQualificationTimeoutExpires();
            break;    
        case EVENT_ANNOUNCE_RECEIPT_TIMEOUT_EXPIRES:
            eventAnnounceReceiptTimeoutExpires(eventIn);
            break;    
        case EVENT_PERIODIC_ANNOUNCE:
            eventPeriodicAnnounce();
            break;
        case EVENT_PERIODIC_SYNC:
            eventPeriodicSync();
            break;
        case EVENT_DELAY_REQ:
            eventDelayReq();
            break;
        case EVENT_FOLLOW_UP:
            eventFollowUp();
            break;
        
        // Currently not supported externally.
        /*
        case EVENT_INITIALIZE:
            eventInitialize();
            break
        case EVENT_DESIGNATED_ENABLED:
            eventDesignatedEnabled();
            break;
        case EVENT_DESIGNATED_DISABLED:
            eventDesignatedDisabled();
            break;
        case EVENT_FAULT_CLEARED:
            eventFaultCleared();
            break;    
        case EVENT_FAULT_DETECTED:
            eventFaultDetected();
            break;            
        case EVENT_STATE_DECISION_EVENT:
            eventStateDecisionEvent();
            break;
        case EVENT_SYNCHRONIZATION_FAULT:
            eventSynchronizationFault();
            break;
        case EVENT_MASTER_CLOCK_SELECTED:
            eventMasterClockSelected();
            break;
        */
    }

}

/** Use the specifications in 8.2 and the values specified in ptp-config.h to
 * initialize the data sets in PTPState \a ptp.
 */
void PTPInitialize() {
    
    // Set up data structures and default values.
    ClockIdentity clockIdentity = CLOCKIDENTITY;

    // DefaultDS Initialization (8.2.1).
    ptp.defaultDS.twoStepFlag = TWOSTEPFLAG;
    memcpy(ptp.defaultDS.clockIdentity, clockIdentity, 8);
    ptp.defaultDS.numberPorts = NUMBERPORTS;
    if(SLAVEONLY) {
        ptp.defaultDS.clockQuality.clockClass = 255;
    }
    else {
        ptp.defaultDS.clockQuality.clockClass = CLOCKCLASS;
    }
    ptp.defaultDS.clockQuality.clockAccuracy = CLOCKACCURACY;
    ptp.defaultDS.clockQuality.offsetScaledLogVariance = 
        OFFSETSCALEDLOGVARIANCE;
    ptp.defaultDS.priority1 = PRIORITY1;
    ptp.defaultDS.priority2 = PRIORITY2;
    ptp.defaultDS.domainNumber = DOMAINNUMBER;
    ptp.defaultDS.slaveOnly = SLAVEONLY;
    
    // CurrentDS Initialization (8.2.2).
    ptp.currentDS.stepsRemoved = 0;
    ptp.currentDS.offsetFromMaster.scaledNanoseconds = 0;
    ptp.currentDS.meanPathDelay.scaledNanoseconds = MEANPATHDELAY;
    ptp.currentDS.lastOffsetFromMaster.scaledNanoseconds = 0;

    // ParentDS Initialization (8.2.3).
    memcpy(ptp.parentDS.parentPortIdentity.clockIdentity, 
           ptp.defaultDS.clockIdentity, 8);
    ptp.parentDS.parentPortIdentity.portNumber = 0;
    ptp.parentDS.parentStats = FALSE;
    ptp.parentDS.observedParentOffsetScaledLogVariance = 0xFFFF;
    ptp.parentDS.observedParentClockPhaseChangeRate = 0x7FFFFFFF;
    memcpy(ptp.parentDS.grandmasterIdentity, ptp.defaultDS.clockIdentity, 8);
    ptp.parentDS.grandmasterClockQuality = ptp.defaultDS.clockQuality; 
    ptp.parentDS.priority1 = ptp.defaultDS.priority1;
    ptp.parentDS.priority2 = ptp.defaultDS.priority2;
    
    // TimeProperties Initialization (8.2.4).
    ptp.timePropertiesDS.currentUtcOffset = CURRENTUTCOFFSET; 
    ptp.timePropertiesDS.currentUtcOffsetValid = CURRENTUTCOFFSETVALID;
    ptp.timePropertiesDS.leap59 = LEAP59;
    ptp.timePropertiesDS.leap61 = LEAP61;
    ptp.timePropertiesDS.timeTraceable = TIMETRACEABLE;
    ptp.timePropertiesDS.frequencyTraceable = FREQUENCYTRACEABLE;
    ptp.timePropertiesDS.ptpTimescale = PTPTIMESCALE;
    ptp.timePropertiesDS.timeSource = TIMESOURCE;

    // PortDS Initialization (8.2.5).
    ptp.portDS.portIdentity.portNumber = 1;
    memcpy(ptp.portDS.portIdentity.clockIdentity, clockIdentity, 8);
    ptp.portDS.portState = INITIALIZING;
    ptp.portDS.logMinDelayReqInterval = LOGMINDELAYREQINTERVAL;
    ptp.portDS.peerMeanPathDelay.scaledNanoseconds = 0;
    ptp.portDS.logAnnounceInterval = LOGANNOUNCEINTERVAL;
    ptp.portDS.announceReceiptTimeout = ANNOUNCERECEIPTTIMEOUT;
    ptp.portDS.logSyncInterval = LOGSYNCINTERVAL;
    ptp.portDS.delayMechanism = E2E;
    ptp.portDS.versionNumber = VERSIONPTP;
    
    // SequenceId Initialization.
    ptp.sequenceIdSync = 0;
    ptp.sequenceIdDelayReq = 0;
    ptp.sequenceIdAnnounce = 0;
    ptp.sequenceIdManagement = 0;

    // Synchronization algorithm initialization.
    ptp.syncEventTimeInterval.scaledNanoseconds = 0;
    ptp.delayEventTimeInterval.scaledNanoseconds = 0;
    ptp.lastCorrectionField = 0;
    ptp.rateAdjFreq = 0;
    ptp.rateAdjFreqAcc = 0;
    ptp.rateAdjOffset = 0;
    ptp.rateAdjOffsetAcc = 0;

    // This allows old timeout events to be ignored when they fire the actor.
    ptp.announceReceiptTimeoutExpiresID = 0;
    
    // Current index for foreignMasterDS.
    ptp.foreignMasterDSIndex = 0;
    
    ptp.sameMaster = TRUE;
    
    // Start periodic announce interval.  This periodic firing is still
    // used when in slave mode and Announce messages are not being sent.
    eventPeriodicAnnounce();
    
}

/** Calculate the mean path delay using the Delay_Resp message and the latest 
 * Sync (or Follow_Up) message. Since clock adjustment is performed after a 
 * Sync message and clock adjustment affects the mean path delay calculation
 * (by affecting the timestamps of messages), take those changes into account.
 * This code can be modified to add filtering to deal with the variability of
 * mean path delay.
 */
void computeMeanPathDelay() {
    
    // Represent difference in timestamps as a time interval 
    // (scaled nanoseconds as defined by standard).
    toTimeInterval(&ptp.delayEventEgressTimestamp,
                   &ptp.delayEventIngressTimestamp,
                   &ptp.delayEventTimeInterval);

    // Use correction field from Delay_Resp message.
    ptp.delayEventTimeInterval.scaledNanoseconds -=  ptp.lastCorrectionField;
    
    // Calculate meanPathDelay.
    ptp.currentDS.meanPathDelay.scaledNanoseconds = 
        (ptp.syncEventTimeInterval.scaledNanoseconds >> 1) + 
        (ptp.delayEventTimeInterval.scaledNanoseconds >> 1); 
    
    // Take into account clock changes between Sync Event and Delay Event.
    // Correct for clock step or rate adjust to correct offset 
    // (currently ignores rate adjust to correct drift).
    if(ptp.lastAdjType == STEP || ptp.lastAdjType == STEP_2) {
        ptp.currentDS.meanPathDelay.scaledNanoseconds -= 
            (ptp.currentDS.offsetFromMaster.scaledNanoseconds >> 1);
    } else if(ptp.lastAdjType == RATE) {
    }

}

/** Calculate the offset from master using the difference in timestamps for 
 * the egress and ingress sync events and the current value of mean path delay.
 */
void computeOffsetFromMaster() {

    // Save last offsetFromMaster.
    ptp.currentDS.lastOffsetFromMaster.scaledNanoseconds =
        ptp.currentDS.offsetFromMaster.scaledNanoseconds;

    // Represent difference in timestamps as a time interval 
    // (scaled nanoseconds as defined by standard).
    toTimeInterval(&ptp.syncEventEgressTimestamp,
            &ptp.syncEventIngressTimestamp,
            &ptp.syncEventTimeInterval);

    // Use correction field from Sync (or Follow_Up) message.
    ptp.syncEventTimeInterval.scaledNanoseconds -=  ptp.lastCorrectionField;

    // Calculate offsetFromMaster.
    ptp.currentDS.offsetFromMaster.scaledNanoseconds =
        ptp.syncEventTimeInterval.scaledNanoseconds - 
        ptp.currentDS.meanPathDelay.scaledNanoseconds;
    
}

/** How the clock is adjusted depends on the amount of adjustment required.
 * If the offset from master is outside the cutoff, the clock is stepped since
 * it would take too long to syncrhonize with frequency adjustment.  If it is
 * within the cutoff, a rateAdjFreq is calcuated to attempt to make the
 * frequency of the slave clock equal to that of the master. Even if the 
 * frequencies are equal, an offset may still exist (even though it wouldn't
 * change if the frequencies are equal). To account for this, a temporary
 * rateAdjOffset is also calculated to attempt to reduce the offset
 * by the next time adjustClock() is called. There is a PI controller for 
 * adjusting rateAdjFreq and a PI controller for adjusting rateAdjOffset. 
 * This code can be modified to change the servo algorithm. 
 */

void adjustClock() {
    
    Integer64 offset, freqAdj, offsetAdj;
    
    // Check if within cutoff.
    // Get offset in ns.
    offset = (ptp.currentDS.offsetFromMaster.scaledNanoseconds >> 16);
    if((offset > CUTOFF) || (offset < -CUTOFF)) {

        // Slave clock is too far seperated from master clock to be able to 
        // syncrhronize solely with frequency adjustments, so step the clock.
        stepClockTimeInterval(&ptp.currentDS.offsetFromMaster);

        // Fix timestamps.
        fixTimestamps(&ptp.currentDS.offsetFromMaster);

        // Set adjustment type.
        ptp.lastAdjType = STEP;

    } else {
           
        // Offset (in ns).
        offsetAdj = offset;

        // Change in offset (in ns).
        if(ptp.lastAdjType == STEP || ptp.lastAdjType == STEP_2) {
            // Clock was stepped so change in offset not valid.
            freqAdj = offsetAdj;
        } else {
            freqAdj = offset -
                (ptp.currentDS.lastOffsetFromMaster.scaledNanoseconds >> 16);
        }
        
        // Convert to units of 100ps/s.
        // Use the time until next Sync message.
        if(ptp.portDS.logSyncInterval > 0) {
            offsetAdj = (-10*offsetAdj) >> ptp.portDS.logSyncInterval;
            freqAdj = (-10*freqAdj) >> ptp.portDS.logSyncInterval;
        } else {
            offsetAdj = (-10*offsetAdj) << (-ptp.portDS.logSyncInterval);
            freqAdj = (-10*freqAdj) << (-ptp.portDS.logSyncInterval);
        }
        
        if(ptp.lastAdjType == STEP) {

            // When clock is stepped, the mean path delay is most likely not 
            // correctly known until after (For example, when adjustClock()
            // is first called).

            // Now step the clock again with the valid delay.
            stepClockTimeInterval(&ptp.currentDS.offsetFromMaster);

            // Fix timestamps.
            fixTimestamps(&ptp.currentDS.offsetFromMaster);

            // Set adjustment rates.
            ptp.rateAdjFreq = freqAdj;
            ptp.rateAdjFreqAcc = 0;
            ptp.rateAdjOffset = 0;
            ptp.rateAdjOffsetAcc = 0;
            rateAdjustClock(ptp.rateAdjFreq);

            // Set adjustment type.
            ptp.lastAdjType = STEP_2;
        
        } else {

            // rateAdjOffset is used as a temporary rate adjustment to fix 
            // offset, so remove this effect from change in offset since last 
            // sync message.
            freqAdj += ptp.rateAdjOffset;

            // Since last synchronization, any change in offset should be due 
            // solely to clock drift (After effect of rateAdjOffset is 
            // corrected). rateAdjFreq is updated to attempt to 
            // adjust frequency to lock to master clock. Even if this 
            // synchronizes the clock rate of master and slave, there will still
            // be an offset, so rateAdjOffset temporarily adjusts the rate to 
            // correct for this offset by the time of the next adjustClock().
            
            // Set adjustment rates.
            ptp.rateAdjFreq += FREQ_P(freqAdj);
            ptp.rateAdjFreqAcc += FREQ_I(freqAdj);
            ptp.rateAdjOffset = OFFSET_P(offsetAdj);
            ptp.rateAdjOffsetAcc = OFFSET_I(offsetAdj);
            
            rateAdjustClock(ptp.rateAdjFreq + ptp.rateAdjFreqAcc + 
                            ptp.rateAdjOffset + ptp.rateAdjOffsetAcc);

            // Set adjustement type.
            ptp.lastAdjType = RATE;
            
       }
        
    }
    
    printStatus();

}

/** Create an event of type 'EVENT_ANNOUNCE_RECEIPT_TIMEOUT_EXPIRES' with a
 * timestamp of 'current time + announce interval'.  Call addFutureEvent()
 * with this event to have this event fire PTPActor() when current time
 * reaches that of the timestamp.
 */
void generateAnnounceReceiptTimeout() {
    
    Event event;

    //DEBUG_PRINTF("Announce Receipt Timeout\n");

    // This will cause all past announceReceiptTimeouts to be ignored.
    ptp.announceReceiptTimeoutExpiresID++;
    
    // Set correct event timestamp.
    getClock(&event.eventTimestamp);
    //TODO: currently doesn't add random number (9.2.6.11)
    (event.eventTimestamp.secondsField.lsb) +=
        ptp.portDS.announceReceiptTimeout * 
        (1 << ptp.portDS.logAnnounceInterval);
    
    // Assign event type.
    event.eventType = EVENT_ANNOUNCE_RECEIPT_TIMEOUT_EXPIRES;
    
    // Assign ID to event.
    event.eventDataPointer = (void *)malloc(sizeof(UInteger16));
    *((UInteger16 *)event.eventDataPointer) = 
        ptp.announceReceiptTimeoutExpiresID; 

    // Add event.
    addFutureEvent(&event);    
    
}

/** Create an event of type 'EVENT_QUALIFICATION_TIMEOUT_EXPIRES' with a
 * timestamp of 'current time + N * announce interval'.  Call addFutureEvent()
 * with this event to have this event fire PTPActor() when current time
 * reaches that of the timestamp.
 */
void generateQualificationTimeout(UInteger8 N) {
    
    Event event;
    
    // Set correct event timestamp.
    getClock(&event.eventTimestamp);
    (event.eventTimestamp.secondsField.lsb) +=
        N * (1 << ptp.portDS.logAnnounceInterval);
        
    // Assign event type.
    event.eventType = EVENT_QUALIFICATION_TIMEOUT_EXPIRES;

    // Add event.
    addFutureEvent(&event);
        
}

/** Handle PTP messages by calling the corresponding message handling function.
 */
void eventMsgIn(const Event *eventIn) {

    UInteger8* msg = (UInteger8 *)eventIn->eventDataPointer;
    UInteger4 messageType = msg[0] & 0x0F;
    
    switch(messageType) {

        case SYNC:
            syncMsgIn(msg);
            break;
        case FOLLOW_UP:
            followUpMsgIn(msg);
            break;
        case DELAY_REQ:
            delayReqMsgIn(msg);
            break;
        case DELAY_RESP:
            delayRespMsgIn(msg);
            break;
        case ANNOUNCE:
            announceMsgIn(msg);
            break;
        case MANAGEMENT:
            managementMsgIn(msg);

    }

}

/** Perform state transition as specified in 9.2.5. */
void eventPowerup() {
   
    // Figure 23/24 (9.2.5).
    ptp.portDS.portState = INITIALIZING;
    stateInitializing();
    
}

/** Perform state transition as specified in 9.2.5. */
void eventInitialize() {
    
    // Figure 23/24 (9.2.5).
    ptp.portDS.portState = INITIALIZING;
    stateInitializing();
    
}

/** Perform state transition as specified in 9.2.5. */
void eventDesignatedEnabled() {
    
    // Figure 23/24 (9.2.5).
    if(ptp.portDS.portState == DISABLED) {
        ptp.portDS.portState = INITIALIZING;
        stateInitializing();
    }
}

/** Perform state transition as specified in 9.2.5. */
void eventDesignatedDisabled() {
    
    // Figure 23/24 (9.2.5).
    if(ptp.portDS.portState == LISTENING ||
       ptp.portDS.portState == UNCALIBRATED ||
       ptp.portDS.portState == SLAVE ||
       ptp.portDS.portState == PRE_MASTER ||
       ptp.portDS.portState == MASTER ||
       ptp.portDS.portState == PASSIVE ||
       ptp.portDS.portState == FAULTY) {
                   
        ptp.portDS.portState = DISABLED;
        stateDisabled();
    }
     
}

/** Perform state transition as specified in 9.2.5. */
void eventFaultCleared() {
    
    // Figure 23/24 (9.2.5).
    if(ptp.portDS.portState == FAULTY) {
        ptp.portDS.portState = INITIALIZING;
        stateInitializing();
    }
     
}

/** Perform state transition as specified in 9.2.5. */
void eventFaultDetected() {
    
    // Figure 23/24 (9.2.5).
    if(ptp.portDS.portState == LISTENING ||
       ptp.portDS.portState == UNCALIBRATED ||
       ptp.portDS.portState == SLAVE ||
       ptp.portDS.portState == PRE_MASTER ||
       ptp.portDS.portState == MASTER ||
       ptp.portDS.portState == PASSIVE ||
       ptp.portDS.portState == FAULTY) {
                   
        ptp.portDS.portState = FAULTY;
        stateFaulty();
                        
    }
                
}

/** Run the Best master clock (BMC) algorithm as specified in 9.3.  Depending 
 * on the result of the BMC algorithm, perform state transition as specified in 
 * 9.2.5. */
void eventStateDecisionEvent() {
    
    UInteger8 i,j;
    Integer8 best;
    ClockDataSet a,b;
    
    // Cannot occur in INITIALIZING state.
    if(ptp.portDS.portState == INITIALIZING)
        return;

    // Figure 25 (9.2.6.8).
            
    // Find Erbest (9.3.2.3).
    best = -1;

    // Check if any foreign clocks have been detected.
    if(ptp.foreignMasterDSIndex > 0) {
                
        // Find first foreign master with qualified announce message.
        for(i = 0; i < ptp.foreignMasterDSIndex; i++) {
                    
            // Check if announce message is qualified (9.3.2.5).
            if(isQualified(&ptp.foreignMasterDS[i]) == TRUE) {
                toDataSetFromMsg(&a, &ptp.foreignMasterDS[i].msg);
                best = i;
                break;
            }
                     
        }
                
        // Compare against any remaining foreign masters with qualified
        // announce messages.
        for(i = i + 1; i < ptp.foreignMasterDSIndex; i++) {
                    
            // Check if announce message is qualified (9.3.2.5).
            if(isQualified(&ptp.foreignMasterDS[i]) == TRUE) {
                toDataSetFromMsg(&b, &ptp.foreignMasterDS[i].msg);
                
                // Compare clocks.
                j = compareDataSet(&a, &b);

                // Check if b is better than a.
                if(j == DATA_SET_B || j == DATA_SET_B_T) {
                    a = b;
                    best = i;    
                }
                        
            }
                     
        }
                
    }
        
    // At this point, 'best' contains the index value of the best 
    // foreign master (or -1 if none exist), and 'a' the clock info.
    if(best != -1) {
        
        // Make 'best' the only foreign master in data set (9.3.2.3b).
        if(best != 0) {        
            ptp.foreignMasterDS[0] = ptp.foreignMasterDS[best];    
        }
        ptp.foreignMasterDSIndex = 1;
        
    }
            
    // Figure 26 (9.3.3).
    if(best == -1 && ptp.portDS.portState == LISTENING)
        return;
                
    // Local clock.            
    toDataSetFromLocal(&b);
                
    if(ptp.defaultDS.clockQuality.clockClass <= 127) {
                
        if(best == -1) {
            
            // BMC_MASTER (D0) M1
            updateDataSet(STATE_DECISION_CODE_M1, NULL);
            i = BMC_MASTER;
            
        } else {
            
            j = compareDataSet(&a, &b);
            
            if(j == DATA_SET_B || j == DATA_SET_B_T) {
            
                // BMC_MASTER (D0) M1
                updateDataSet(STATE_DECISION_CODE_M1, NULL);
                i = BMC_MASTER;
            
            } else {
            
                // BMC_PASSIVE (Erbest) P1
                updateDataSet(STATE_DECISION_CODE_P1, NULL);
                i = BMC_PASSIVE;
            
            }
            
        }
    
    } else {
    
        if(best == -1) {
    
            // BMC_MASTER (D0) M2
            updateDataSet(STATE_DECISION_CODE_M2, NULL);
            i = BMC_MASTER;
    
        } else {
    
            j = compareDataSet(&a,&b);
    
            if(j == DATA_SET_B || j == DATA_SET_B_T) {
                        
                // BMC_MASTER (D0) M2
                updateDataSet(STATE_DECISION_CODE_M2, NULL);
                i = BMC_MASTER;
                
            } else {
            
                // BMC_SLAVE
                updateDataSet(STATE_DECISION_CODE_S1,
                              &ptp.foreignMasterDS[best].msg);
                i = BMC_SLAVE;
                        
            }
            
        }
        
    }
            
    // Update state (Figure 23/24) (9.2.5).
    if((ptp.portDS.portState == LISTENING ||
       ptp.portDS.portState == UNCALIBRATED ||
       ptp.portDS.portState == SLAVE ||
       ptp.portDS.portState == PRE_MASTER ||
       ptp.portDS.portState == PASSIVE) &&
       i == BMC_MASTER && 
       ptp.defaultDS.slaveOnly == FALSE) {
                   
        ptp.portDS.portState = PRE_MASTER;
        statePreMaster();
        return;
                
    }
             
    if((ptp.portDS.portState == LISTENING ||
       ptp.portDS.portState == UNCALIBRATED ||
       ptp.portDS.portState == PRE_MASTER ||
       ptp.portDS.portState == MASTER ||
       ptp.portDS.portState == PASSIVE) &&
       i == BMC_SLAVE) {
                   
        ptp.portDS.portState = UNCALIBRATED;
        stateUncalibrated();
        return;
                
    }
            
    if((ptp.portDS.portState == LISTENING ||
       ptp.portDS.portState == UNCALIBRATED ||
       ptp.portDS.portState == SLAVE ||
       ptp.portDS.portState == PRE_MASTER ||
       ptp.portDS.portState == MASTER ||
       ptp.portDS.portState == PASSIVE) &&
       i == BMC_PASSIVE &&
       ptp.defaultDS.slaveOnly == FALSE) {
                   
        ptp.portDS.portState = PASSIVE;
        statePassive();
        return;
                
    }
        
    if(ptp.portDS.portState == SLAVE && 
       i == BMC_SLAVE && ptp.sameMaster == FALSE) {
        
        ptp.portDS.portState = UNCALIBRATED;
        stateUncalibrated();
        return;        
        
    }                  
            
    if(ptp.defaultDS.slaveOnly == TRUE &&
      (ptp.portDS.portState == LISTENING ||
       ptp.portDS.portState == UNCALIBRATED ||
       ptp.portDS.portState == SLAVE) &&
      (i == BMC_MASTER || i == BMC_PASSIVE)) {
    
        ptp.portDS.portState = LISTENING;
        stateListening();
               
    }
}

/** Perform state transition as specified in 9.2.5. */
void eventQualificationTimeoutExpires() {
    
    // Figure 23/24 (9.2.5).
    if(ptp.portDS.portState == PRE_MASTER) {
                
        ptp.portDS.portState = MASTER;
        stateMaster();
                    
    }        
    
}

/** If the event is still relevent, perform state transition as specified
 * in 9.2.5.
 */
void eventAnnounceReceiptTimeoutExpires(const Event *eventIn) {
    
    // Figure 23/24 (9.2.5).
    UInteger16 Id = *((UInteger16 *)eventIn->eventDataPointer);

    // Free space allocated to storing announceReceiptTimeoutExpiresID.
    free(eventIn->eventDataPointer);

    // Check if event is still relevant.
    if(Id == ptp.announceReceiptTimeoutExpiresID) {
                            
        if(ptp.portDS.portState == LISTENING ||
           ptp.portDS.portState == UNCALIBRATED ||
           ptp.portDS.portState == SLAVE ||
           ptp.portDS.portState == PASSIVE) {
          
                   
               if(ptp.defaultDS.slaveOnly == TRUE) {
                
                ptp.portDS.portState = LISTENING;
                stateListening();
                return;
                
            }
                  
            // Update port's data sets to MASTER. (9.2.6.11)
            updateDataSet(STATE_DECISION_CODE_M1, NULL);
            ptp.portDS.portState = MASTER;
            stateMaster();
                
        } else {
            
            // Restart AnnouceReceiptTimeout. (b)
            generateAnnounceReceiptTimeout();
            
        }
                        
    }
                        
}

/** Perform state transition as specified in 9.2.5. */
void eventSynchronizationFault() {
    
    // Figure 23/24.
    if(ptp.portDS.portState == SLAVE) {
               
        ptp.portDS.portState = UNCALIBRATED;
        stateUncalibrated();
                    
    }
             
}

/** Perform state transition as specified in 9.2.5. */
void eventMasterClockSelected() {
    
    // Figure 23/24.
    if(ptp.portDS.portState == UNCALIBRATED) {
               
        ptp.portDS.portState = SLAVE;
        stateSlave();
                    
    }
            
}

/** Create an event to fire PTPActor() at 'announce interval' in the future,
 * and trigger the BMC algorithm. If master, will also send announce message.
 */
void eventPeriodicAnnounce() {
        
    Event eventOut;
    
    if(ptp.portDS.portState == MASTER) {
                
        // Send Announce message.
        sendAnnounceMsg(&ptp);    
                            
    }
    
    // Set time of next Announce message.
    getClock(&eventOut.eventTimestamp);
    (eventOut.eventTimestamp.secondsField.lsb) +=
    (1 << ptp.portDS.logAnnounceInterval);
    eventOut.eventType = EVENT_PERIODIC_ANNOUNCE;
    addFutureEvent(&eventOut);
    
    // Run BMC algorithm.
    eventStateDecisionEvent();
    
    printStatus();
    
}

/** Create an event to fire the PTPActor at 'sync interval' in the future. 
 * If master, also send a sync message (and follow_up in 2 step mode).
 */
void eventPeriodicSync() {
    
    Event eventOut;
    TimeInterval timeInterval;
    
    if(ptp.portDS.portState == MASTER) {
               
        // Send Sync message.
        sendSyncMsg(&ptp);
        
        getClock(&eventOut.eventTimestamp);
        
        // Schedule Follow_Up if required.
        if(ptp.defaultDS.twoStepFlag == TRUE) {
            
            eventOut.eventType = EVENT_FOLLOW_UP;
            addFutureEvent(&eventOut);
                
        }
                
        // Set time of next Sync message.
        
        // Set initially to 1s.
        timeInterval.scaledNanoseconds = 1000000000;
        timeInterval.scaledNanoseconds = timeInterval.scaledNanoseconds << 16;
        
        // Modify 1s to correct time based on logSyncInterval.
        if(ptp.portDS.logSyncInterval > 0) {
            timeInterval.scaledNanoseconds = 
                timeInterval.scaledNanoseconds << ptp.portDS.logSyncInterval;   
        } else {
            timeInterval.scaledNanoseconds = 
                timeInterval.scaledNanoseconds >> (-ptp.portDS.logSyncInterval);
        }

        addTimeIntervalToTimestamp(&timeInterval, &eventOut.eventTimestamp);
        eventOut.eventType = EVENT_PERIODIC_SYNC;
        addFutureEvent(&eventOut);
        
    }
    
}

/** Send a Delay_Resp message. */
void eventDelayReq() {
    
    // Send Delay Req message.
    sendDelayReqMsg(&ptp);
    
}

/** Send a Follow_Up message. */
void eventFollowUp() {
    
    // Get timestamp of Sync message.
    getSyncEventEgressTimestamp(&ptp);
    
    // Send Follow Up message.
    sendFollowUpMsg(&ptp);
    
}

/** Initialize data sets, stop AnnounceReceiptTimeout(9.2.6.11) and transition 
 * to LISTENING state. 
 */
void stateInitializing() {
    
    //DEBUG_PRINTF("State: INITIALIZING\n");

    // Initialize data sets.
    PTPInitialize();
            
    // Stop AnnounceReceiptTimeout (9.2.6.11).
    ptp.announceReceiptTimeoutExpiresID++;
            
    // Now that the port has been initialized, 
    // transition to LISTENING state. (9.2.5, Fig 23/24)
    ptp.portDS.portState = LISTENING;
    stateListening();

}

/** Stop AnnounceReceiptTimeout (9.2.6.11). */
void stateFaulty() {
    
    //DEBUG_PRINTF("State: FAULTY\n");

    // Stop AnnounceReceiptTimeout (9.2.6.11).
    ptp.announceReceiptTimeoutExpiresID++;
    
}

/** Stop AnnounceReceiptTimeout (9.2.6.11). */
void stateDisabled() {
    
    //DEBUG_PRINTF("State: DISABLED\n");

    // Stop AnnounceReceiptTimeout (9.2.6.11).
    ptp.announceReceiptTimeoutExpiresID++;
    
}

/** Restart AnnounceReceiptTimeout (9.2.6.11c). */
void stateListening() {

    //DEBUG_PRINTF("State: LISTENING\n");

    // Restart AnnouceReceiptTimeout (9.2.6.11c). 
    generateAnnounceReceiptTimeout();
    
}

/** Stop AnnounceReceiptTimeout (9.2.6.11) and start QualificationTimeout. */
void statePreMaster() {

    //DEBUG_PRINTF("State: PRE_MASTER\n");
    
    // Stop AnnounceReceiptTimeout.
    ptp.announceReceiptTimeoutExpiresID++;
            
    // Start QualificationTimeout.
    // M3 not supported (9.2.6.10)
    generateQualificationTimeout(0);
            
}

/** Stop AnnounceReceiptTimeout (9.2.6.11) and start periodically sending
 * sync messages. */
void stateMaster() {
   
   //DEBUG_PRINTF("State: MASTER\n");

    // Stop AnnounceReceiptTimeout.
    ptp.announceReceiptTimeoutExpiresID++;
    
    // Periodically send sync message.
    eventPeriodicSync();
                  
}

/** Restart AnnounceReceiptTimeout (9.2.6.11c). */
void statePassive() {
    
    //DEBUG_PRINTF("State: PASSIVE\n");

    // Restart AnnouceReceiptTimeout (9.2.6.11c).
    generateAnnounceReceiptTimeout();
            
}

/** Restart AnnounceReceiptTimeout (9.2.6.11c). */
void stateUncalibrated() {
    
    //DEBUG_PRINTF("State: UNCALIBRATED\n");

    // Restart AnnouceReceiptTimeout (9.2.6.11c).
    generateAnnounceReceiptTimeout();
            
    eventMasterClockSelected();

}

/** Restart AnnounceReceiptTimeout (9.2.6.11c). */
void stateSlave() {
    
    //DEBUG_PRINTF("State: SLAVE\n");

    // Restart AnnouceReceiptTimeout (9.2.6.11c).
    generateAnnounceReceiptTimeout();
    
}

/** Process the Announce message as specified in 9.5.3. */
void announceMsgIn(UInteger8* msg) {
    
    AnnounceMsg announceMsg;
    UInteger8 i;
    
    // Receipt of Announce message logic (9.5.3, Figure 29).
    if(ptp.portDS.portState == INITIALIZING ||
       ptp.portDS.portState == DISABLED ||
       ptp.portDS.portState == FAULTY) {
        return;
    }
             
    // Parse the received Announce message.
    readAnnounceMsg(&announceMsg);
        
    // Break if alternateMasterFlag is True.
    if((announceMsg.header.flagField[0] & 0x01) == TRUE) {
        return;
    }
            
    //TODO: announce receipt timeout (d) (PASSIVE state) (9.2.6.11)
            
    // Check if from current master.
    if(comparePortIdentity(&announceMsg.header.sourcePortIdentity, 
                           &ptp.parentDS.parentPortIdentity) == TRUE) {
        
        if(ptp.portDS.portState == UNCALIBRATED || 
           ptp.portDS.portState == SLAVE) {
                
            // Restart AnnouceReceiptTimeout. (a)
            generateAnnounceReceiptTimeout();

        }

        //TODO: if SLAVE, update data sets?
    }
    
    // Add or update foreignMasterDS.
    for(i = 0; i < ptp.foreignMasterDSIndex; i++) {

        if(comparePortIdentity(&announceMsg.header.sourcePortIdentity, 
           &ptp.foreignMasterDS[i].foreignMasterPortIdentity) == TRUE) {
 
            // Already exists.
            
            // Not qualified if not most recent.
            if(announceMsg.header.sequenceId < 
                    ptp.foreignMasterDS[i].msg.header.sequenceId) {
                break;
            }
            
            // Update timestamps of announce messages.
            ptp.foreignMasterDS[i].announceMessageTimestamps[0] = 
                ptp.foreignMasterDS[i].announceMessageTimestamps[1];
            getClock(&ptp.foreignMasterDS[i].announceMessageTimestamps[1]);
                    
            // Update to latest announce message.
            ptp.foreignMasterDS[i].msg = announceMsg;
                                        
            break;
                       
        }
    }
    
    if(i == ptp.foreignMasterDSIndex) {
                
        // Create new record.
                
        // Prevent overflow of foreignMasterDS.
        if(ptp.foreignMasterDSIndex >= FOREIGNMASTERDSSIZE) {
            ptp.foreignMasterDSIndex = FOREIGNMASTERDSSIZE - 1;
        }
                
        // Add to foreignMasterDS.
        // Initialize.
        ptp.foreignMasterDS[ptp.foreignMasterDSIndex].
            foreignMasterPortIdentity = announceMsg.header.sourcePortIdentity;
        ptp.foreignMasterDS[ptp.foreignMasterDSIndex].
            foreignMasterAnnounceMessages = 0;
        ptp.foreignMasterDS[ptp.foreignMasterDSIndex].
            msg = announceMsg;
        ptp.foreignMasterDS[ptp.foreignMasterDSIndex].
            announceMessageTimestamps[0].secondsField.msb = 0;
        ptp.foreignMasterDS[ptp.foreignMasterDSIndex].
            announceMessageTimestamps[0].secondsField.lsb = 0;
        ptp.foreignMasterDS[ptp.foreignMasterDSIndex].
            announceMessageTimestamps[0].nanosecondsField = 0;
        getClock(&ptp.foreignMasterDS[ptp.foreignMasterDSIndex].
                 announceMessageTimestamps[1]);
        ptp.foreignMasterDSIndex++;
            
    }

}

/** Process the Sync message as specified in 9.5.4. */
void syncMsgIn(UInteger8* msg) {
    
    SyncMsg syncMsg;
    Event eventOut;
    Integer64 ns;
    TimeInterval timeInterval;
    
    // Receipt of Sync message logic (9.5.4, Figure 30).
    if(ptp.portDS.portState == INITIALIZING ||
       ptp.portDS.portState == DISABLED ||
       ptp.portDS.portState == FAULTY ||
       !(ptp.portDS.portState == SLAVE || 
         ptp.portDS.portState == UNCALIBRATED)) {
        clearIngressTimestamp();
        return;
    }
             
    // Parse the received Sync message.
    readSyncMsg(&syncMsg);
            
    // Make sure it is from current master clock.
    if(comparePortIdentity(&syncMsg.header.sourcePortIdentity,
                &ptp.parentDS.parentPortIdentity) == FALSE) {
        clearIngressTimestamp();
        return;
    }
            
    // Get syncEventIngressTimestamp.
    if(getSyncEventIngressTimestamp(&ptp, syncMsg.header.sequenceId) == FALSE) {
        return;
    }
    
    
    // Set time to send Delay_Req (9.5.11.2).
    getClock(&eventOut.eventTimestamp);
    ns = ((ptp.portDS.logMinDelayReqInterval + 1) << 1) * (1000000000);
    ns = (ns * (rand() % ((1 << DELAYREQDIV) + 1))) >> DELAYREQDIV;
    timeInterval.scaledNanoseconds = ns << 16;
    addTimeIntervalToTimestamp(&timeInterval, &eventOut.eventTimestamp);
    eventOut.eventType = EVENT_DELAY_REQ;
    addFutureEvent(&eventOut);
        
    // Syncronize local clock if twoStepFlag is false.
    if((syncMsg.header.flagField[0] & 0x02) == 0) {
        
        // Accurate timestamp is in Sync Message.
        ptp.syncEventEgressTimestamp = syncMsg.originTimestamp;
        
        // Get Correction Field
        ptp.lastCorrectionField = syncMsg.header.correctionField 
                + DELAY_ASYMMETRY;
        
        // Calculate offsetFromMaster.
        computeOffsetFromMaster();
        
        // Synchronize!
        adjustClock();
                
    }     
    
    // Store sync message information.
    ptp.lastSyncMsgSourcePortIdentity = syncMsg.header.sourcePortIdentity;
    ptp.lastSyncMsgSequenceId = syncMsg.header.sequenceId;
    ptp.lastSyncMsgCorrectionField = syncMsg.header.correctionField;
    
}

/** Process the Follow_Up message as specified in 9.5.5. */
void followUpMsgIn(UInteger8* msg) {
    
    FollowUpMsg followUpMsg;
    
    // Receipt of Follow_Up message logic (9.5.5, Figure 31).
    if(ptp.portDS.portState == INITIALIZING ||
       ptp.portDS.portState == DISABLED ||
       ptp.portDS.portState == FAULTY ||
       !(ptp.portDS.portState == SLAVE || 
         ptp.portDS.portState == UNCALIBRATED)) {

        return;
    }
             
    // Parse the received Follow_Up message.
    readFollowUpMsg(&followUpMsg);
            
    // Make sure it is from current master clock.
    if(comparePortIdentity(&followUpMsg.header.sourcePortIdentity,
                           &ptp.parentDS.parentPortIdentity) == FALSE) {
        return;
    }
            
    // Make sure it is associated with the last sync message.
    if(comparePortIdentity(&followUpMsg.header.sourcePortIdentity,
                           &ptp.lastSyncMsgSourcePortIdentity) == TRUE &&
       followUpMsg.header.sequenceId == ptp.lastSyncMsgSequenceId) {
        
        // Store accurate timestamp for sync message.
        ptp.syncEventEgressTimestamp = followUpMsg.preciseOriginTimestamp;
        
        // Get Correction Field
        ptp.lastCorrectionField = ptp.lastSyncMsgCorrectionField + 
                                  followUpMsg.header.correctionField + 
                                  DELAY_ASYMMETRY;
        
        // Calculate offsetFromMaster.
        computeOffsetFromMaster();
        
        // Synchronize!
        adjustClock();
        
    }

}

/** Process the Delay_Req message as specified in 9.5.6. */
void delayReqMsgIn(UInteger8* msg) {
    
    DelayReqMsg delayReqMsg;
    
    // Receipt of Delay_Req message logic (9.5.6, Figure 32).
    if(ptp.portDS.portState == INITIALIZING ||
       ptp.portDS.portState == DISABLED ||
       ptp.portDS.portState == FAULTY ||
       ptp.portDS.portState != MASTER) {
        clearIngressTimestamp();
        return;
    }
    
    // Parse the received Delay Req message.    
    readDelayReqMsg(&delayReqMsg);  
    
    // Get timestamp.
    if(getDelayEventIngressTimestamp(&ptp, delayReqMsg.header.sequenceId) 
            == FALSE) {
        return;
    }  

    // Send Delay Resp message.
    sendDelayRespMsg(&ptp, &delayReqMsg);
        
}

/** Process the Delay_Resp message as specified in 9.5.6. */
void delayRespMsgIn(UInteger8* msg) {
    
    DelayRespMsg delayRespMsg;
    
    // Receipt of Delay_Resp message logic (Figure 33).
    if(ptp.portDS.portState == INITIALIZING ||
       ptp.portDS.portState == DISABLED ||
       ptp.portDS.portState == FAULTY ||
       !(ptp.portDS.portState == SLAVE || 
         ptp.portDS.portState == UNCALIBRATED)) {
        return;
    }
    
    // Parse the received Delay Resp mesage.         
    readDelayRespMsg(&delayRespMsg);
            
    // Check if Delay_Resp message is associated with most recent Delay_Req 
    // message.
    if((comparePortIdentity(&delayRespMsg.header.sourcePortIdentity,
                           &ptp.parentDS.parentPortIdentity) == FALSE) ||
       (comparePortIdentity(&delayRespMsg.requestingPortIdentity,
                           &ptp.portDS.portIdentity) == FALSE) ||
       (delayRespMsg.header.sequenceId != ptp.sequenceIdDelayReq)) {
        return;
    }
        
    // Get Delay Req message egress timestamp.
    getDelayEventEgressTimestamp(&ptp);
    
    // Get Delay Req message ingress timestamp from Delay Resp message.
    ptp.delayEventIngressTimestamp = delayRespMsg.receiveTimestamp;
    
    // Update
    ptp.portDS.logMinDelayReqInterval = delayRespMsg.header.logMessageInterval;
    
    // Get Correction Field
    ptp.lastCorrectionField = delayRespMsg.header.correctionField;
        
    // Calculate meanPathDelay.
    computeMeanPathDelay();
            
}

/** Process the Management message as specified in 15. */
void managementMsgIn(UInteger8* msg){
    
    ManagementMsg managementMsg;
    
    readManagementMsg(&managementMsg);

    //TODO: Implement management messages.
    
}

/** Fill the ClockSetSet \a ds common data set with pointers to values in the
 * Announce message (9.3.4).
 */ 
void toDataSetFromMsg(ClockDataSet *ds, AnnounceMsg *msg) {

    ds->priority1 = &msg->grandmasterPriority1;
    ds->identity = &msg->grandmasterIdentity;
    ds->clockClass = &msg->grandmasterClockQuality.clockClass;
    ds->accuracy = &msg->grandmasterClockQuality.clockAccuracy;
    ds->offsetScaledLogVariance = 
        &msg->grandmasterClockQuality.offsetScaledLogVariance;
    ds->priority2 = &msg->grandmasterPriority2;
    ds->stepsRemoved = &msg->stepsRemoved;
    ds->sender = &msg->header.sourcePortIdentity.clockIdentity;
    ds->receiver = &ptp.parentDS.parentPortIdentity.clockIdentity; 
    ds->receiverPortNumber = &ptp.parentDS.parentPortIdentity.portNumber; 
    
}

/** Fill the ClockSetSet \a ds common data set with pointers to values in the
 * local data sets (9.3.4).
 */ 
void toDataSetFromLocal(ClockDataSet *ds) {
    
    ds->priority1 = &ptp.defaultDS.priority1;
    ds->identity = &ptp.defaultDS.clockIdentity;
    ds->clockClass = &ptp.defaultDS.clockQuality.clockClass;
    ds->accuracy = &ptp.defaultDS.clockQuality.clockAccuracy;
    ds->offsetScaledLogVariance = 
        &ptp.defaultDS.clockQuality.offsetScaledLogVariance;
    ds->priority2 = &ptp.defaultDS.priority2;
    ds->stepsRemoved = &zero;
    ds->sender = &ptp.defaultDS.clockIdentity;
    ds->receiver = &ptp.defaultDS.clockIdentity; 
    ds->receiverPortNumber = &zero;

    
}

/** Use the data comparison algorithm specified in 9.3.4 to compare two data
 * sets, and return result of comparison.
 */
UInteger8 compareDataSet(const ClockDataSet *a, const ClockDataSet *b) {
    
    UInteger8 i;
    
    // Data comparison algorithm as in Figure 27 (9.3.4).
    i = memcmp(*a->identity, *b->identity, 8);
    if(i == 0) {
        // Data comparison algorithm as in Figure 28 (9.3.4).
        if(*a->stepsRemoved > *b->stepsRemoved + 1) {
            return DATA_SET_B;
        } else if(*a->stepsRemoved + 1 < *b->stepsRemoved) {
            return DATA_SET_A;
        } else if(*a->stepsRemoved > *b->stepsRemoved) {
            i = memcmp(*a->receiver, *a->sender, 8);
            if(i == 0) {
                return DATA_SET_ERROR_1;
            } else if (*a->receiver[i] < *a->sender[i]) {
                return DATA_SET_B;
            } else {
                return DATA_SET_B_T;
            }
        } else if(*a->stepsRemoved < *b->stepsRemoved) {
            i = memcmp(*b->receiver, *b->sender, 8);
            if(i == 0) {
                return DATA_SET_ERROR_1;
            } else if (*b->receiver[i] < *b->sender[i]) {
                return DATA_SET_A;
            } else {
                return DATA_SET_A_T;
            }
        } else {
            i = memcmp(*a->sender, *b->sender, 8);
            if(i == 0) {
                if(*a->receiverPortNumber > *b->receiverPortNumber) {
                    return DATA_SET_B_T;
                } else if(*a->receiverPortNumber < *b->receiverPortNumber) {
                    return DATA_SET_A_T;
                } else {
                    return DATA_SET_ERROR_2;
                }    
            } else if(*a->sender[i] > *b->sender[i]) {
                return DATA_SET_B_T;
            } else {
                return DATA_SET_A_T;
            }
        }
    } else if(*a->priority1 > *b->priority1) {
        return DATA_SET_B;
    } else if(*a->priority1 < *b->priority1) {
        return DATA_SET_A;
    } else if(*a->clockClass > *b->clockClass) {
        return DATA_SET_B;
    } else if(*a->clockClass < *b->clockClass) {
        return DATA_SET_A;
    } else if(*a->accuracy > *b->accuracy) {
        return DATA_SET_B;
    } else if(*a->accuracy < *b->accuracy) {
        return DATA_SET_A;
    } else if(*a->offsetScaledLogVariance > *b->offsetScaledLogVariance) {
        return DATA_SET_B;
    } else if(*a->offsetScaledLogVariance < *b->offsetScaledLogVariance) {
        return DATA_SET_A;
    } else if(*a->priority2 > *b->priority2) {
        return DATA_SET_B;
    } else if(*a->priority2 < *b->priority2) {
        return DATA_SET_A;
    } else if(*a->identity[i] > *b->identity[i]) {
        return DATA_SET_B;
    } else {
        return DATA_SET_A;
    }
    
}

/** Update data sets as specified in 9.3.5. */
void updateDataSet(UInteger8 code, const AnnounceMsg *msg) {

    switch(code) {
        
        case STATE_DECISION_CODE_M1:
        case STATE_DECISION_CODE_M2:
        
            // Table 13 (.3.5).
            ptp.currentDS.stepsRemoved = 0;    
            ptp.currentDS.offsetFromMaster.scaledNanoseconds = 0;
            ptp.currentDS.meanPathDelay.scaledNanoseconds = 0;
            memcpy(ptp.parentDS.parentPortIdentity.
                   clockIdentity,ptp.defaultDS.clockIdentity, 8);
            ptp.parentDS.parentPortIdentity.portNumber = 0;
            memcpy(ptp.parentDS.grandmasterIdentity,
                   ptp.defaultDS.clockIdentity,8);
            ptp.parentDS.grandmasterClockQuality = ptp.defaultDS.clockQuality;
            ptp.parentDS.priority1 = ptp.defaultDS.priority1;
            ptp.parentDS.priority2 = ptp.defaultDS.priority2;
            // TODO: timePropertiesDS update (9.4). 
            /*
            ptp.timePropertiesDS.currentUtcOffset =
            ptp.timePropertiesDS.currentUtcOffsetValid = 
            ptp.timePropertiesDS.leap59 = 
            ptp.timePropertiesDS.leap61 = 
            ptp.timePropertiesDS.timeTraceable = 
            ptp.timePropertiesDS.frequencyTraceable = 
            ptp.timePropertiesDS.ptpTimescale = 
            ptp.timePropertiesDS.timeSource =
            */
            break;
            
        case STATE_DECISION_CODE_M3:
            break;
            
        case STATE_DECISION_CODE_S1:
        
            // Table 16.
            ptp.currentDS.stepsRemoved = 1 + msg->stepsRemoved;    
            if(comparePortIdentity(&ptp.parentDS.parentPortIdentity,
                                   &msg->header.sourcePortIdentity) == TRUE) {
                ptp.sameMaster = TRUE;
            } else {
                ptp.sameMaster = FALSE;
                memcpy(ptp.parentDS.parentPortIdentity.clockIdentity,
                       msg->header.sourcePortIdentity.clockIdentity, 8);
            }
            ptp.parentDS.parentPortIdentity.portNumber = 
                msg->header.sourcePortIdentity.portNumber;
            memcpy(ptp.parentDS.grandmasterIdentity,
                   msg->grandmasterIdentity, 8);
            ptp.parentDS.grandmasterClockQuality = msg->grandmasterClockQuality;
            ptp.parentDS.priority1 = msg->grandmasterPriority1;
            ptp.parentDS.priority2 = msg->grandmasterPriority2;
            ptp.timePropertiesDS.currentUtcOffset = msg->currentUtcOffset;
            ptp.timePropertiesDS.currentUtcOffsetValid = 
                (msg->header.flagField[1] & 0x3);
            ptp.timePropertiesDS.leap59 = (msg->header.flagField[1] & 0x2);
            ptp.timePropertiesDS.leap61 = (msg->header.flagField[1] & 0x1);
            ptp.timePropertiesDS.timeTraceable = 
                (msg->header.flagField[1] & 0x5);
            ptp.timePropertiesDS.frequencyTraceable = 
                (msg->header.flagField[1] & 0x6);
            ptp.timePropertiesDS.ptpTimescale = 
                (msg->header.flagField[1] & 0x4);
            ptp.timePropertiesDS.timeSource = msg->timeSource;
            break;
            
        case STATE_DECISION_CODE_P1:
            break;
        case STATE_DECISION_CODE_P2:
            break;    
    }    
    
}

/** If the PortIdentity.clockIdentity and PortIdentity.portNumber of two
 * PortIdentities are both equal, then the PortIdentities are equal.
 */
Boolean comparePortIdentity(const PortIdentity *a, const PortIdentity *b) {
    
    if(memcmp(a->clockIdentity, b->clockIdentity, 8) == 0 && 
       a->portNumber == b->portNumber)
        return TRUE;
    else
        return FALSE;
            
}

/** Check if an Announce message is qualified as specified in 9.3.2.5. */
Boolean isQualified(const ForeignMasterDS *d) {
    
    Integer64 timeWindow;
    TimeInterval timeInterval;
    Timestamp timestamp;
    getClock(&timestamp);
    
    if(d->msg.stepsRemoved >= 255) {
        return FALSE;
    }

    // Find difference between current time and 2nd to last announce message.
    toTimeInterval(&d->announceMessageTimestamps[0], &timestamp, &timeInterval);
    
    // Change back to ns.
    timeInterval.scaledNanoseconds = timeInterval.scaledNanoseconds >> 16;
    
    // Calcuate time window in ns.
    timeWindow = 1000000000;
    timeWindow = timeWindow * 
        (FOREIGN_MASTER_TIME_WINDOW * (1 << ptp.portDS.logAnnounceInterval));
    
    // Check if within FOREIGN_MASTER_TIME_WINDOW announce intervals.    
    if(timeInterval.scaledNanoseconds < timeWindow) {
        return TRUE;
    } else {
        return FALSE;
    }
    
}

/** Fix the timestamps of when Announce messages were received in the 
 * foreignMasterDS records.
 */
void fixTimestamps(const TimeInterval *timeInterval) {

    UInteger8 i;
    TimeInterval t;
    t.scaledNanoseconds = -timeInterval->scaledNanoseconds;
    
    for(i = 0; i < ptp.foreignMasterDSIndex; i++) {
    
        addTimeIntervalToTimestamp(&t, 
                &ptp.foreignMasterDS[i].announceMessageTimestamps[0]);
        addTimeIntervalToTimestamp(&t, 
                &ptp.foreignMasterDS[i].announceMessageTimestamps[1]);
               
    }

}

/** Print the state of PTP along with offset from master (in SLAVE state)
 * to LCD display.
 */
void printStatus() {
    
    char line1[9];
    char line2[9];
    Integer32 o;
    
    if(ptp.portDS.portState == 6) {
        sprintf(line1, "Master  ");
        sprintf(line2, "        ");
    } else if(ptp.portDS.portState == 9) {
        o = ptp.currentDS.offsetFromMaster.scaledNanoseconds >> 16;
        sprintf(line1, "Slave   ");
        sprintf(line2, "%d", o);
    } else {
        sprintf(line1, "%d      ",ptp.portDS.portState);
        sprintf(line2, "        ");
    }
    
    DisplayClear();
    DisplayString(0, line1);
    DisplayString(16, line2);
 
}
